"""Entry point for the CLI tools."""

# 1. Standard Python modules

# 2. Third party modules

# 3. Aquaveo modules

# 4. Local modules
from harmonica.cli import main

if __name__ == "__main__":
    main.main()
